// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import BigInt from 'big-integer';

export default class BaseScalar {
    /**
     * @param {(bigInt.BigInteger|number)} value
     * @param {bigInt.BigInteger} p
     */
    constructor(value, p) {
        this.value = typeof value === 'number' ? BigInt(value) : value;
        this.p = p;
    }

    /**
     * @param {(BaseScalar|bigInt.BigInteger)} scalar
     */
    plus(scalar) {
        return this.add(scalar);
    }

    /**
     * Performs addition
     * @param {(BaseScalar|bigInt.BigInteger)} scalar
     */
    add(scalar) {
        return new BaseScalar(this.value.add(getValue(scalar)).mod(this.p), this.p); 
    }

    /**
     * @param {(BaseScalar|bigInt.BigInteger)} scalar
     */
    multiply(scalar) {
        return this.times(scalar);
    }

    /**
     * @param {(BaseScalar|bigInt.BigInteger)} scalar
     */
    times(scalar) {
        return new BaseScalar(this.value.multiply(getValue(scalar)).mod(this.p), this.p); 
    }

    negate() {
        return new BaseScalar(this.p.subtract(this.value), this.p); 
    }
    
    /**
     * @param {(BaseScalar|bigInt.BigInteger)} scalar
     */
    subtract(scalar) {
        return this.minus(scalar);
    }

    /**
     * @param {(BaseScalar|bigInt.BigInteger)} scalar
     */
    minus(scalar) {
        return new BaseScalar(this.value.subtract(getValue(scalar)).mod(this.p), this.p);
    }

    toString(radix = 10) {
        return this.value.toString(radix);
    }

    /**
     * @returns {number[]}
     */
    toArray() {
        return this.value.toArray(256).value;
    }

    inspect() {
        return this.toString(16);
    }
}
/**
 * 
 * @param {(BaseScalar|bigInt.BigInteger)} scalar
 * @returns {(bigInt.BigInteger)}
 */
function getValue(scalar) {
    return scalar instanceof BaseScalar ? scalar.value : scalar;
}