// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import elliptic from 'elliptic';
import utils from './utils';
import EcScalar from './ecScalar';
import BigInt from 'big-integer';

const BasePoint = elliptic.curve.base.BasePoint;
const curve = (new elliptic.ec('secp256k1')).curve;

export default class EcPoint {
    static get infinity() { return new EcPoint(curve.point(null, null)); }
    static get g() { return new EcPoint(curve.g); }

    constructor(x, y) {
        /**
         *@type {elliptic.curve.base.BasePoint}
         */
        this._value = x instanceof BasePoint ? x
            : Number.isInteger(x) && Number.isInteger(y) ? curve.point(x, y) 
            : curve.point(x.toString(16), y.toString(16));
    }

    get x() {
        return BigInt.fromArray(this._value.getX().toArray(true), 256, false);
    }

    get y() {
        return BigInt.fromArray(this._value.getY().toArray(true), 256, false);
    }

    get isInfinity() {
        return this._value.isInfinity();
    }

    get isValid() {
        return this._value.validate();
    }

    plus(point) {
        return this.add(point);
    }

    add(point) {
        return new EcPoint(this._value.add(point._value));
    }

    multiply(num) {
        return this.times(num);
    }

    /**
     * @param {(EcScalar|BigInt|number)} num
     */
    times(num) {
        // this.value = typeof value === 'number' ? BigInt(value) : value;
        var value = (num instanceof EcScalar || num instanceof BigInt) ? num.toString(16) : num;
        return new EcPoint(this._value.mul(value));
    }

    negate() {
        return new EcPoint(this._value.neg());
    }
    
    subtract(point) {
        return this.minus(point);
    }

    minus(point) {
        return this.add(point.negate());
    }

    values() {
        return [this.x, this.y];
    }

    toArray() {
        return utils.padLeft(this._value.getX().toArray(true), 32).concat(utils.padLeft(this._value.getY().toArray(true), 32));
    }

    inspect() {
        return `<point x: ${this._value.getX().toString(16)} y: ${this._value.getY().toString(16)} >`;
    }
}

