// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import assert from 'assert';
import BigInt from 'big-integer';
import createHash from 'create-hash';
import secretShare from './secretShare.js';
import utils, { constants } from './utils.js';
import Hash from './hash.js';

export default class ElGamalKey {
    /**
     * 
     * @param {boolean} privated 
     * @param {bigInt.BigInteger} p
     * @param {bigInt.BigInteger} q
     * @param {bigInt.BigInteger} g
     * @param {bigInt.BigInteger} key
     * @param {bigInt.BigInteger} id
     */
    constructor(privated, p, q, g, key, id = undefined) {
        this.privated = privated;
        this.p = p;
        this.q = q;
        this.g = g;
        this.key = key; //represent the secret in elgammal and the y in secret sharing
        this.id = id; //represent the x value in secret sharing
    }

    get size() {
        return this.p.bitLength().valueOf();
    }

    createPublic() {
        const key = this.g.modPow(this.key, this.p);
        return new ElGamalKey(false, this.p, this.q, this.g, key);
    }

    share(number, threshold) {
        const [xs, ys] = secretShare.share(this.key, number, threshold, this.q);
        return ys.map((yi, i) => new ElGamalKey(this.privated, this.p, this.q, this.g, yi, xs[i]));
    }

    static combine(shares) {
        const first = shares[0];
        const [ids, keys] = [shares.map(sh => sh.id), shares.map(sh => sh.key)]
        const key = secretShare.interpolate(ids, keys, first.q);

        return new ElGamalKey(first.privated, first.p, first.q, first.g, key)
    }

    toString() {
        const values = [this.p, this.g, this.key];
        if (this.id)
            values.push(this.id);

        const bytes = utils.encodeByteArray(values, this.size, true);
        var checksum = Hash.shaBuffer(bytes)[0];
        const falgs = ((this.privated ? 1 : 0) << 1 | (this.id ? 1 : 0)) << 6 | (checksum >> 2);

        bytes.unshift(Math.ceil(this.size / 8));
        bytes.unshift(falgs);
        
        return Buffer.from(bytes).toString('base64');
    }

    static from(text) {
        const [flags, size,  ...bytes] = Buffer.from(text, 'base64');
        const privated = (flags & 0b10000000) > 0;
        const numbers = (flags & 0b1000000) ? 4 : 3;
        var checksum = Hash.shaBuffer(bytes)[0];

        assert(bytes.length == numbers * size && checksum >> 2 === (flags & 63) , 'Malformed Key');

        const [p, g, key, id ] = utils.decode(bytes, bytes.length * 8 / numbers);
        const q = p.minus(constants.one).divide(constants.two);

        return new ElGamalKey(privated, p, q, g, key, id);
    }

    static generate(bits = 256, confidence = 32) {
        const [p, q] = utils.randomPrime(bits, confidence);
        const g = utils.getPrimitiveRoot(p).modPow(constants.two, p);
        
        const x = utils.random(constants.one, q.subtract(constants.one));
        const y = g.modPow(x, p);

        return [new ElGamalKey(true, p, q, g, x), new ElGamalKey(false, p, q, g, y)];
    }

    //TODO: Improve this
    static seed(pwd) {
        const g = BigInt('f3e5bf98c73a188850b10f5c7525f4336e81b361db4a92bf43c4404c32b7e7a', 16);
        const p = BigInt('8cec1ccc497f729aa968af8b379a61e967591301b60d0a5b05504f301c93db97', 16);
        const q = p.minus(constants.one).divide(constants.two);

        const bytes = Math.ceil(p.bitLength().valueOf() / 8);
        const x = seedPrivate(pwd, bytes);
        const y = g.modPow(x, p);

        return [new ElGamalKey(true, p, q, g, x), new ElGamalKey(false, p, q, g, y)];
    }
}

function seedPrivate(pwd, bytes) {
    const hashes = Array(Math.ceil(bytes / 32));
    for (let i = 0; i < hashes.length; i++) {
        const data = i ? hashes[i - 1] : pwd;
        hashes[i] = createHash('sha256').update(data).digest();
    }

    var hash = Buffer.concat(hashes).slice(-bytes);
    return hash.reduce((rtn, num) => rtn.shiftLeft(8).or(num), BigInt.zero)
}