// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import EcScalar from './ecScalar';
import TSignClient, { TResponseAuth, TSignGetClient } from './tsignClient';
import ecSecretShare from './ecSecretShare';
import ElGamalKey from './elGamalKey';
import ElGamal from './elGamal';

export default class TAuthFlow {
    
    /**
     * @param {string[]} urls
     * @param {string} user
     */
    constructor(urls, user, isPost = false) {
        /**
        * @type {TSignClient[]}
        */
        this.clients = urls.map(url => new (isPost ? TSignClient : TSignGetClient)(url, user));
        this.user = user;
    }

    /**
     * @param {EcScalar} key
     * @param {EcScalar} pass
     */
    async signIn(key, pass) {
        try {
            var keyShrs = await getShares(this.clients, key);
            var passShrs = await getShares(this.clients, pass);
            await Promise.all(this.clients.map((cln, i) => cln.SetAccount(passShrs[i], keyShrs[i])));
        } catch (err) {
            return Promise.reject(err);
        }
    }

    /**
     * @param {EcScalar} password
     */
    async logIn(password) {
        try {
            var rs = await Promise.all(this.clients.map(cln => cln.getAuthRi()));
            var ais = await getShares(this.clients, password);

            var alphaPromises = Promise.all(this.clients.map((cln, i) => cln.getAuthAlpha(ais[i].value, rs)));
            var [sec, pub] = ElGamalKey.generate(128);
            var alphas = await alphaPromises;
            
            var toIdx = Array.from(rs.keys()).map(j => rs.map((_, i) => i).filter(i => i !== j));
            var allAlphas = alphas.map((alpha, i) => alpha.public.map((data, j) => ({ i: toIdx[i][j], data: data }))).reduce((a, b) => a.concat(b));
            
            var xis = await Promise.all(this.clients.map((cln, i) => cln.getAuthSi(alphas[i].private, pub, allAlphas.filter(alpha => i === alpha.i).map(({data}) => data))));
            
            var priv = xis.map(cypher => TResponseAuth.from(ElGamal.decrypt(cypher.priv, sec))).reduce((sum, xtagi) => sum.add(xtagi.si), EcScalar.zero);
            
            var key = { priv: new ElGamalKey(true, xis[0].pub.p, xis[0].pub.q, xis[0].pub.g, priv.value), pub: xis[0].pub };
            
            
            return key;
        } catch (err) {
            return Promise.reject(err);
        }
    }
}


/**
 * @param {TSignClient[]} clients
 * @param {EcScalar} password
 */
async function getShares(clients, password) {
    var ids = await Promise.all(clients.map(cln => cln.getId()));
    var [_, shrs] = ecSecretShare.shareFromIds(password, ids.map(id => new EcScalar(id)), ids.length);
    return shrs;
}
