// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import EcPoint from './ecPoint';
import bitcoin from './bitcoin';
import BigInt from 'big-integer';
import EcScalar from './ecScalar';
import TSignClient from './tsignClient';
import ecSecretShare from './ecSecretShare';
import { TSignGetClient } from './tsignClient';

export default class TSignFlow {
    
    /**
     * @param {string[]} urls
     * @param {string} user
     */
    constructor(urls, user, isPost = false) {
        /**
        * @type {TSignClient[]}
        */
        this.clients = urls.map(url => new (isPost ? TSignClient : TSignGetClient)(url, user));
        this.user = user;
        this.onProgress = (id, msg) => {};
    }

    /**
     * @param {EcScalar} key
     */
    async share(key) {
        var ids = await Promise.all(this.clients.map(cln => cln.getId()));
        var [_, shrs] = ecSecretShare.shareFromIds(key, ids.map(id => new EcScalar(id)), ids.length);
        await Promise.all(this.clients.map((cln, i) => cln.SetAccount(EcScalar.zero, shrs[i])));
    }

    async signFromValue(m) {
        try {
            this.onProgress(0, "starting");
            var ks = await Promise.all(this.clients.map((cln) => cln.getSignKi(m)));
            var alphaArgs = Array.from(ks.keys()).map(j => ks.map((ki, i) => ({ i, ki })).filter((_, i) => i !== j));
    
            this.onProgress(1, "processing");
            var alphas = await Promise.all(this.clients.map((cln, i) => cln.getSignAlpha(alphaArgs[i].map(arg => arg.ki))));
            var allAlphas = alphas.map((resp, i) => resp.public.map((alpha, j) => ({ j: alphaArgs[i][j].i, alpha }))) //zip alphas and is
                .reduce((alphas, alpha) => alphas.concat(alpha));
    
            this.onProgress(2, "processing");
            var deltas = await Promise.all(this.clients.map((cln, i) => cln.getSignDelta(alphas[i].private, allAlphas.filter(({ j, _ }) => i == j).map(({ _, alpha }) => alpha))));
            var allDeltas = deltas.map(resp => resp.public);
            
            this.onProgress(3, "processing");
            var signs = await Promise.all(this.clients.map((cln, i) => cln.getSignRSi(deltas[i].private, allDeltas)));
            var s = signs.reduce((sum, sign) => sum.add(sign.si), EcScalar.zero);
            //EcScalar
    
            this.onProgress(4, "finished");
            var y = signs.map(shri => new EcPoint(shri.xi, shri.yi)).reduce((sum, yi) => sum.add(yi), EcPoint.infinity);
            return [signs[0].r, s.value, y];
        } catch (err) {
            this.onProgress(4, "error");
            return Promise.reject(err);
        }
    }

    async sign(message) {
        var m = getM(message);
        var [r, s, y] = await this.signFromValue(m);
        
        var address = bitcoin.getAddress(y);
        var signature = bitcoin.formatSign(m, r, s, address);

        return { address, signature };    
    }
}

function getM(message) { 
    return BigInt.isInstance(message) ? message : bitcoin.getM(message);
}